import math
import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm

# import celloracle as co
CONFIG = {
    "default_args": {"lw": 0.3, "rasterized": True},
    "s_scatter": 5,
    "s_grid": 20,
    "scale_simulation": 30,
    "scale_dev": 30,
    "cmap_ps": "PiYG",
    "default_args_quiver": {"linewidths": 0.25, "width": 0.004},
}


def get_predecessors(g, links):
    return links[links["target"] == g]["source"].unique()


def get_upstream_genes(target_gene, links, num_hops=3):
    upstream_genes = set()
    new_genes = [target_gene]
    for i in range(num_hops):
        curr_genes = new_genes
        new_genes = []
        for g in curr_genes:
            preds = get_predecessors(g, links)
            for pred in preds:
                if pred not in upstream_genes:
                    upstream_genes.add(pred)
                    new_genes.append(pred)
    return upstream_genes


def get_descendents(g, links):
    return links[links["source"] == g]["target"].unique()


def add_signs(links):
    links["sign"] = np.zeros(links.shape[0], dtype=int)
    links.loc[links["coef_mean"] > 0, "sign"] = 1
    links.loc[links["coef_mean"] < 0, "sign"] = -1

    return links


def get_signs(source, links, targets):
    signs = []
    coefs = []
    source_links = links[links["source"] == source]
    target_links = source_links[source_links["target"].isin(targets)]
    return list(target_links.loc[:, "sign"]), list(target_links.loc[:, "coef_mean"])


def get_downstream_genes(target_gene, links, num_hops=3, use_coefs=True):
    # This could change to also return the coefficients from the dictionary and do some kind of calculation on them.
    # Maybe just start with the sign
    downstream_genes = set()
    downstream_genes_signs = {}
    downstream_genes_signs[target_gene] = 1
    if use_coefs == True:
        downstream_genes_coefs = {}
        downstream_genes_coefs[target_gene] = math.log(1)
    new_genes = [target_gene]
    add_signs(links)
    for i in range(num_hops):
        curr_genes = new_genes
        new_genes = []
        for g in curr_genes:
            descs = get_descendents(g, links)
            signs, coefs = get_signs(g, links, descs)
            for j, desc in enumerate(descs):
                if desc not in downstream_genes:
                    downstream_genes.add(desc)
                    new_genes.append(desc)
                    if use_coefs == True:
                        """If coefs[j] != 0.0: downstream_genes_coefs[desc] =
                        downstream_genes_signs[g] + math.log(abs(coefs[j]))

                        else:
                            downstream_genes_coefs[desc] = downstream_genes_signs[g] - 800

                        """
                        downstream_genes_coefs[desc] = coefs[j]
                    downstream_genes_signs[desc] = downstream_genes_signs[g] * signs[j]
                else:
                    if (
                        downstream_genes_signs[desc]
                        != downstream_genes_signs[g] * signs[j]
                    ):
                        print(
                            f"for gene {desc}, it has a different sign coming from {g} than from previous. It has a coef of {coefs[j]}.",
                        )
    for t in downstream_genes_signs:
        downstream_genes_signs[t] = (
            downstream_genes_signs[t] * downstream_genes_coefs[t]
        )
    return downstream_genes, downstream_genes_signs


def get_metagenes_score(
    adata,
    regulators,
    m_name,
    links,
    links_dict_num="0",
    negative_signs=True,
    use_coefs=True,
):
    regdf = pd.DataFrame(
        index=regulators,
        columns=adata.uns["spicemix_genes"].astype(str),
        dtype=int,
    )
    regdf.loc[:, :] = 0
    for regnum, reg in enumerate(regulators):
        dsg, signs = get_downstream_genes(
            reg,
            links.links_dict[links_dict_num],
            num_hops=1,
            use_coefs=use_coefs,
        )
        for g in dsg:
            if g in regdf.columns:
                if negative_signs == False:
                    regdf.loc[reg, g] = abs(signs[g])
                else:
                    regdf.loc[reg, g] = signs[g]
    regdf = regdf.loc[regdf.sum(axis=1) != 0]
    metagenes = adata.uns["M"][m_name]
    mval_nonorm = np.matmul(regdf, metagenes).astype(float)
    # normalize
    mval = mval_nonorm.div(mval_nonorm.abs().sum(axis=1), axis=0).astype(float)
    return mval, mval_nonorm


def get_intersecting(ms_nonorm, metagene_scores, k=10):
    intersect_lists = []
    top_percent_lists = []
    negative_intersect_lists = []
    negative_top_percent_lists = []
    for i in range(len(metagene_scores.columns)):
        topk = ms_nonorm.loc[ms_nonorm.loc[:, i].sort_values()[-k:].index, :]
        topknorm = metagene_scores.loc[
            metagene_scores.loc[:, i].sort_values()[-k:].index,
            :,
        ]
        intersect = set(topk.index).intersection(set(topknorm.index))
        intersect_lists.append(list(intersect))
        top_percent_lists.append(list(topknorm.index)[-2:])

        negtopk = ms_nonorm.loc[ms_nonorm.loc[:, i].sort_values()[:k].index, :]
        negtopknorm = metagene_scores.loc[
            metagene_scores.loc[:, i].sort_values()[:k].index,
            :,
        ]
        negintersect = set(negtopk.index).intersection(set(negtopknorm.index))
        negative_intersect_lists.append(list(negintersect))
        negative_top_percent_lists.append(list(negtopknorm.index)[:2])
    return (
        top_percent_lists,
        intersect_lists,
        negative_top_percent_lists,
        negative_intersect_lists,
    )


from sklearn.ensemble import BaggingRegressor

# from celloracle.utility import standard, intersect
from sklearn.linear_model import Ridge


# from celloracle.network.regression_models import _get_coef_matrix
def get_edges_window(ad_ex, ad_motif, target_gene, grn, ad_pop, num_hops=1):

    # subset the ad_ex and ad_motif by the neighbors
    tfs = intersect(grn[target_gene], ad_motif.var_names)
    retdf = pd.DataFrame(index=tfs)
    for i, cell in enumerate(ad_ex.obs_names):
        #         neighbors_bool = np.asarray(ad_pop.obsp['adjacency_matrix'][i,:].todense().astype(bool)).flatten()
        neighbors_bool = get_nhop_neighbors(ad_pop, i, num_hops=num_hops)
        neighbors = ad_ex.obs_names[neighbors_bool]
        if len(neighbors) == 0:
            retdf[cell] = np.zeros((len(tfs), 1))
            continue
        data = ad_motif[neighbors, tfs].to_df()
        label = ad_ex[neighbors, target_gene].to_df()
        model = BaggingRegressor(
            base_estimator=Ridge(
                alpha=1,
                solver="auto",
                random_state=123,
            ),
            n_estimators=10,
            bootstrap=True,
            max_features=0.8,
            verbose=False,
            random_state=123,
        )
        model.fit(data, label)
        ans = _get_coef_matrix(model, tfs).mean(axis=0)
        retdf[cell] = ans
    return retdf.T


def get_neighbors(ad_pop, i, nn_indices, cell_type, cluster_id, num_within, num_total):
    # we would like to get a list of cells within cell type and outside cell type ordered by distance
    # We should only do the distance calculations once
    # We could probably use scipy knn for this, then split the list by same or different cell type
    cell_indices = nn_indices[i]
    within_cluster_indices = [
        ind
        for ind, cell in enumerate(ad_pop.obs_names)
        if ad_pop.obs[cluster_id][cell] == cell_type
    ]
    cell_type_list = [t for t in ad_pop.obs[cluster_id].values()]

    num_total += 1  # This is done so you always return yourself plus the number of required neighbors
    neighbors = []
    num_without_added = 0
    cell_idx = 0
    while len(neighbors) < num_total:
        if cell_type_list[cell_indices[cell_idx]] != cell_type:
            if num_without_added < (num_total - num_within):
                neighbors.append(cell_indices[cell_idx])
                num_without_added += 1
        else:
            neighbors.append(cell_indices[cell_idx])
        cell_idx += 1
    # now to get the boolean mask
    neighbors_bool = [
        True if c_ind in neighbors else False for c_ind in range(len(ad_pop.obs_names))
    ]
    return neighbors_bool


def get_nearest_neighbors(ad, n_neighbors=None):
    from sklearn.neighbors import NearestNeighbors

    X = ad.obsm["spatial"]
    if n_neighbors == None:
        n_neighbors = len(ad.obs_names)
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm="ball_tree").fit(X)
    distances, indices = nbrs.kneighbors(X)
    return indices


def get_metagene_edges_window(
    ad_ex,
    ad_motif,
    target_metagene,
    ad_pop,
    num_hops=1,
    cluster_id=None,
    num_within=50,
    num_total=100,
):

    # subset the ad_ex and ad_motif by the neighbors
    tfs = ad_motif.var_names
    retdf = pd.DataFrame(index=tfs)
    if cluster_id != None:
        nn_indices = get_nearest_neighbors(ad_pop)
    for i, cell in tqdm(enumerate(ad_ex.obs_names)):
        #         neighbors_bool = np.asarray(ad_pop.obsp['adjacency_matrix'][i,:].todense().astype(bool)).flatten()
        if cluster_id == None:
            neighbors_bool = get_nhop_neighbors(ad_pop, i, num_hops=num_hops)
        else:
            neighbors_bool = get_neighbors(
                ad_pop,
                i,
                nn_indices,
                ad_ex.obs[cluster_id][cell],
                cluster_id,
                num_within,
                num_total,
            )
        neighbors = ad_ex.obs_names[neighbors_bool]
        if len(neighbors) == 0:
            retdf[cell] = np.zeros((len(tfs), 1))
            continue
        data = ad_motif[neighbors, tfs].to_df()
        # label = ad_pop.obsm['normalized_X'][neighbors_bool,target_metagene]
        label = ad_pop.obsm["X"][neighbors_bool, target_metagene]
        model = BaggingRegressor(
            base_estimator=Ridge(
                alpha=1,
                solver="auto",
                random_state=123,
            ),
            n_estimators=10,
            bootstrap=True,
            max_features=0.8,
            verbose=False,
            random_state=123,
        )
        model.fit(data, label)
        ans = _get_coef_matrix(model, tfs).mean(axis=0)
        retdf[cell] = ans
    return retdf.T


def get_metagene_edges_smoothed(ad_ex, ad_motif, metagene_num, ad_pop, num_hops=1):

    # subset the ad_ex and ad_motif by the neighbors
    tfs = ad_motif.var_names
    retdf = pd.DataFrame(index=tfs)
    for target_metagene in tqdm(range(metagene_num)):
        data = pd.DataFrame(index=tfs, dtype=np.float64)
        label = pd.DataFrame(index=[target_metagene], dtype=np.float64)
        for i, cell in enumerate(ad_ex.obs_names):
            #         neighbors_bool = np.asarray(ad_pop.obsp['adjacency_matrix'][i,:].todense().astype(bool)).flatten()
            neighbors_bool = get_nhop_neighbors(ad_pop, i, num_hops=num_hops)
            neighbors = ad_ex.obs_names[neighbors_bool]
            if len(neighbors) == 0:
                data[cell] = np.zeros((len(tfs), 1))
                label[cell] = 0
                continue
            data[cell] = np.asarray(ad_motif[neighbors, tfs].X.sum(axis=0))
            # label[cell] = np.asarray(ad_pop[neighbors,:].obsm['normalized_X'][:,target_metagene].sum())
            label[cell] = np.asarray(
                ad_pop[neighbors, :].obsm["X"][:, target_metagene].sum(),
            )
        data = data.T - data.T.min(axis=0)
        data /= data.max(axis=0)
        label = label.T - label.T.min()
        label /= label.max()
        model = BaggingRegressor(
            base_estimator=Ridge(
                alpha=1,
                solver="auto",
                random_state=123,
            ),
            n_estimators=100,
            bootstrap=True,
            max_features=0.8,
            verbose=False,
            random_state=123,
        )
        model.fit(data, label)
        ans = _get_coef_matrix(model, tfs).mean(axis=0)
        retdf[str(target_metagene)] = ans
    return retdf


def get_nhop_neighbors(ad, cellidx, num_hops=1):
    neighbors = ad.obsp["adjacency_matrix"][cellidx, :]
    for i in range(num_hops - 1):
        neighbors = neighbors.multiply(ad.obsp["adjacency_matrix"][cellidx, :])
        neighbors[neighbors > 1] = 1
    return np.asarray(neighbors.todense().astype(bool)).flatten()


from multiprocessing import Pool

from popari import Popari, tl
from scipy.stats import zscore


def in_silico_perturb(ad_pop, ad_tf, ad_edge, tf, K=16, multiplier=10, useX=False):
    if useX == False:
        dropout_X = ad_pop.obsm["normalized_X"].copy()
    else:
        dropout_X = ad_pop.obsm["X"].copy()
    # multiply the tf by the edge-weight for each cell for each metagene
    for metagene in range(K):
        perturbation = ad_tf[:, tf].X * ad_edge[:, tf].layers[f"M_{metagene}"]
        perturbation *= multiplier
        dropout_X[:, metagene] = dropout_X[:, metagene] - perturbation.flatten()
    ad_pop.obsm[f"X_{tf}_dropout"] = dropout_X

    normalized_embeddings = zscore(ad_pop.obsm[f"X_{tf}_dropout"])
    nan_mask = np.isnan(normalized_embeddings)
    normalized_embeddings[nan_mask] = 0

    ad_pop.obsm[f"normalized_X_{tf}_dropout"] = normalized_embeddings
    sc.pp.neighbors(ad_pop, use_rep=f"normalized_X_{tf}_dropout")


def align_leiden(d, tf):
    # find the best matching between 'original_leiden' clusters and 'leiden' clusters
    # Can be solved with minimum-weight perfect matching for a bipartite graph of the old and new clusters where the edges are the number of matches between clusters
    # Can use the scipy algorithm for linear_sum_assignment()
    from scipy.optimize import linear_sum_assignment

    # build the cost matrix
    # The issue is when there are more new clusters than there are old clusters. We need to stop this from happening in the function above
    cost_matrix = np.zeros(
        (len(d.obs["original_leiden"].unique()), len(d.obs["leiden"].unique())),
    )
    for cluster in d.obs["leiden"].unique():
        vals = d.obs[d.obs["leiden"] == cluster]["original_leiden"].value_counts()
        inds = d.obs[d.obs["leiden"] == cluster]["original_leiden"].value_counts().index
        for ind, val in zip(inds, vals):
            cost_matrix[int(ind), int(cluster)] = val
    row_matches, col_matches = linear_sum_assignment(cost_matrix, maximize=True)
    remapped_col_matches = [
        row_matches[col_matches.tolist().index(i)] for i in range(len(col_matches))
    ]
    newleiden = [str(remapped_col_matches[int(obs)]) for obs in d.obs["leiden"]]
    # d.obs[f'leiden_{tf}_dropout'] = newleiden
    return newleiden


def test_all_true(l):
    if sum(l) < len(l):
        return False
    else:
        return True


def run_all_perturbations_parallel(
    pop,
    ad_tfs,
    ad_edges,
    K=16,
    multiplier=1,
    useX=False,
    target_clusters=10,
    num_processes=4,
):
    tfs = ad_tfs[0].var_names
    # cluster_changes = {}
    new_columns = [[] for d in pop.datasets]
    checkpoints = 100
    with Pool(processes=num_processes) as pool:
        new_columns.append(
            pool.starmap(
                run_perturbation,
                [(tf, pop, ad_tfs, ad_edges, K, multiplier, useX) for tf in tfs],
            ),
        )

    tfcolumns = [f"leiden_{tf}_dropout" for tf in tfs]
    new_columns_pds = [
        pd.DataFrame(
            new_columns[i],
            index=tfcolumns,
            columns=pop.datasets[i].obs_names,
        ).T
        for i in range(len(pop.datasets))
    ]
    for d, ncpd in zip(pop.datasets, new_columns_pds):
        d.obs = d.obs.join(ncpd)
    # return cluster_changes


def run_perturbation(tf, pop, ad_tfs, ad_edges, K, multiplier, useX):
    for pop_ad, ad_tf, ad_edge in zip(pop.datasets, ad_tfs, ad_edges):
        in_silico_perturb(
            pop_ad,
            ad_tf,
            ad_edge,
            tf,
            K=K,
            multiplier=multiplier,
            useX=useX,
        )
    # tl.leiden(pop, use_rep=f"X_{tf}_dropout", target_clusters=10)
    tl.leiden(
        pop,
        use_rep=f"normalized_X_{tf}_dropout",
        target_clusters=target_clusters,
    )
    j = target_clusters
    len_satisfies = test_all_true(
        [len(d.obs["leiden"].unique()) <= target_clusters for d in pop.datasets],
    )
    while len_satisfies == False:
        j -= 1
        tl.leiden(pop, use_rep=f"normalized_X_{tf}_dropout", target_clusters=j)
        len_satisfies = test_all_true(
            [len(d.obs["leiden"].unique()) <= target_clusters for d in pop.datasets],
        )
    return [align_leiden(d, tf) for d in pop.datasets]


def run_all_perturbations(
    pop,
    ad_tfs,
    ad_edges,
    K=16,
    multiplier=1,
    useX=False,
    target_clusters=10,
    get_leiden=True,
):
    new_columns = [[] for d in pop.datasets]
    for pop_ad, ad_tf, ad_edge, ite in zip(
        pop.datasets,
        ad_tfs,
        ad_edges,
        range(len(pop.datasets)),
    ):
        tfs = ad_tf.var_names
        # cluster_changes = {}
        for it, tf in tqdm(enumerate(tfs)):
            in_silico_perturb(
                pop_ad,
                ad_tf,
                ad_edge,
                tf,
                K=K,
                multiplier=multiplier,
                useX=useX,
            )
        # tl.leiden(pop, use_rep=f"X_{tf}_dropout", target_clusters=10)
        if get_leiden == True:
            tl.leiden(
                pop,
                use_rep=f"normalized_X_{tf}_dropout",
                target_clusters=target_clusters,
            )
            j = target_clusters
            len_satisfies = test_all_true(
                [
                    len(d.obs["leiden"].unique()) <= target_clusters
                    for d in pop.datasets
                ],
            )
            while len_satisfies == False:
                j -= 1
                # tl.leiden(pop, use_rep=f"X_{tf}_dropout", target_clusters=j)
                tl.leiden(pop, use_rep=f"normalized_X_{tf}_dropout", target_clusters=j)
                len_satisfies = test_all_true(
                    [
                        len(d.obs["leiden"].unique()) <= target_clusters
                        for d in pop.datasets
                    ],
                )
            # if the number of clusters is more than 10, change the target number down
            # changes = []
            for tf in tfs:
                new_columns[ite].append(align_leiden(pop_ad, tf))
                # l = (d.obs['original_leiden'] != d.obs[f'leiden_{tf}_dropout']).sum()
                # changes.append(l)
                # d.obs.rename(columns={'new_leiden': f'leiden_{tf}_dropout'})
                # d.obs[f'leiden_{tf}_dropout'] = d.obs['new_leiden']
                # if it % 50 == 0:
                # newd = d.copy()
                # d = newd
            # cluster_changes[tf] = changes
    if get_leiden == True:
        tfcolumns = [f"leiden_{tf}_dropout" for tf in tfs]
        new_columns_pds = [
            pd.DataFrame(
                new_columns[i],
                index=tfcolumns,
                columns=pop.datasets[i].obs_names,
            ).T
            for i in range(len(pop.datasets))
        ]
        for d, ncpd in zip(pop.datasets, new_columns_pds):
            d.obs = d.obs.join(ncpd)
    # return cluster_changes


def find_tf_causing_cluster(pop_ad, cluster_num):
    tf_leidens = [c for c in pop_ad.obs.columns if "dropout" in c]
    changes = {}
    pop_subset = pop_ad[pop_ad.obs["original_leiden"] == cluster_num]
    for tf_leiden in tf_leidens:
        change_num = (pop_subset.obs[tf_leiden] != cluster_num).sum()
        changes[tf_leiden] = change_num
    return changes


#  The functions below were taken from CellOracle (https://github.com/morris-lab/CellOracle)
#  and modified for our use
def _adata_to_matrix(adata, layer_name, transpose=True):
    """Extract an numpy array from adata and returns as numpy matrix.

    Args:
        adata (anndata): anndata

        layer_name (str): name of layer in anndata

        trabspose (bool) : if True, it returns transposed array.

    Returns:
        2d numpy array: numpy array

    """
    if isinstance(adata.layers[layer_name], np.ndarray):
        matrix = adata.layers[layer_name].copy()
    else:
        matrix = adata.layers[layer_name].todense().A.copy()

    if transpose:
        matrix = matrix.transpose()

    return matrix.copy(order="C")


def _obsm_to_matrix(adata, obsm_name, transpose=True):
    """Extract an numpy array from adata and returns as numpy matrix.

    Args:
        adata (anndata): anndata

        layer_name (str): name of layer in anndata

        trabspose (bool) : if True, it returns transposed array.

    Returns:
        2d numpy array: numpy array

    """
    if isinstance(adata.obsm[obsm_name], np.ndarray):
        matrix = adata.obsm[obsm_name].copy()
    else:
        matrix = adata.obsm[obsm_name].todense().A.copy()

    if transpose:
        matrix = matrix.transpose()

    return matrix.copy(order="C")


def plot_background(
    self,
    embedding_name="",
    ax=None,
    s=CONFIG["s_scatter"],
    args=CONFIG["default_args"],
):

    if ax is None:
        ax = plt

    ax.scatter(
        self.embeddings[embedding_name].embedding[:, 0],
        self.embeddings[embedding_name].embedding[:, 1],
        c="lightgray",
        s=s,
        **args,
    )

    # ax.set_title("Pseudotime")
    ax.axis("off")


def plot_cluster_cells_use(
    self,
    embedding_name="",
    ax=None,
    s=CONFIG["s_scatter"],
    color=None,
    show_background=True,
    args=CONFIG["default_args"],
):

    if ax is None:
        ax = plt

    if s == 0:
        color = "white"

    if show_background:
        plot_background(self=self, ax=ax, s=s, args=args)

    if not hasattr(self.embeddings[embedding_name], "cell_idx_use"):
        self.embeddings[embedding_name].cell_idx_use = None

    if self.embeddings[embedding_name].cell_idx_use is None:
        if color is None:
            ax.scatter(
                self.embeddings[embedding_name].embedding[:, 0],
                self.embeddings[embedding_name].embedding[:, 1],
                c=self.colorandum,
                s=s,
                **args,
            )
        else:
            ax.scatter(
                self.embeddings[embedding_name].embedding[:, 0],
                self.embeddings[embedding_name].embedding[:, 1],
                c=color,
                s=s,
                **args,
            )

    else:
        if color is None:
            ax.scatter(
                self.embeddings[embedding_name].embedding[
                    self.embeddings[embedding_name].cell_idx_use,
                    0,
                ],
                self.embeddings[embedding_name].embedding[
                    self.embeddings[embedding_name].cell_idx_use,
                    1,
                ],
                c=self.embeddings[embedding_name].colorandum[
                    self.embeddings[embedding_name].cell_idx_use,
                    :,
                ],
                s=s,
                **args,
            )
        else:
            ax.scatter(
                self.embeddings[embedding_name].embedding[
                    self.embeddings[embedding_name].cell_idx_use,
                    0,
                ],
                self.embeddings[embedding_name].embedding[
                    self.embeddings[embedding_name].cell_idx_use,
                    1,
                ],
                c=color,
                s=s,
                **args,
            )

    ax.axis("off")


def _get_clustercolor_from_anndata(adata, cluster_name, return_as):
    """Extract clor information from adata and returns as palette (pandas data
    frame) or dictionary.

    Args:
        adata (anndata): anndata

        cluster_name (str): cluster name in anndata.obs

        return_as (str) : "palette" or "dict"

    Returns:
        2d numpy array: numpy array

    """

    # return_as: "palette" or "dict"
    def float2rgb8bit(x):
        x = (x * 255).astype("int")
        x = tuple(x)

        return x

    def rgb2hex(rgb):
        return "#%02x%02x%02x" % rgb

    def float2hex(x):
        x = float2rgb8bit(x)
        x = rgb2hex(x)
        return x

    def hex2rgb(c):
        return (int(c[1:3], 16), int(c[3:5], 16), int(c[5:7], 16), 255)

    pal = get_palette(adata, cluster_name)
    if return_as == "palette":
        return pal
    elif return_as == "dict":
        col_dict = {}
        for i in pal.index:
            col_dict[i] = np.array(hex2rgb(pal.loc[i, "palette"])) / 255
        return col_dict
    else:
        raise ValueErroe("return_as")
    return 0


def get_palette(adata, cname):
    c = [i.upper() for i in adata.uns[f"{cname}_colors"]]
    # c = sns.cubehelix_palette(24)
    """
    col = adata.obs[cname].unique()
    col = list(col)
    col.sort()
    """
    try:
        col = adata.obs[cname].cat.categories
        pal = pd.DataFrame({"palette": c}, index=col)
    except:
        col = adata.obs[cname].cat.categories
        c = c[: len(col)]
        pal = pd.DataFrame({"palette": c}, index=col)
    return pal


# this is a function to extract coef information from sklearn ensemble_model.
def _get_coef_matrix(ensemble_model, feature_names):
    # ensemble_model: trained ensemble model. e.g. BaggingRegressor
    # feature_names: list or numpy array of feature names. e.g. feature_names=X_train.columns
    feature_names = np.array(feature_names)
    n_estimater = len(ensemble_model.estimators_features_)
    coef_list = [
        pd.Series(
            ensemble_model.estimators_[i].coef_,
            index=feature_names[ensemble_model.estimators_features_[i]],
        )
        for i in range(n_estimater)
    ]

    coef_df = pd.concat(coef_list, axis=1, sort=False).transpose()

    return coef_df


def intersect(list1, list2):
    """Intersect two list and get components that exists in both list.

    Args:
        list1 (list): input list.
        list2 (list): input list.

    Returns:
        list: intersected list.

    """
    inter_list = list(set(list1).intersection(list2))
    return inter_list



from typing import Optional, Sequence, Union

import matplotlib.pyplot as plt
import scanpy as sc

from .embedding import Embedding
from .eykthyr import Eykthyr


def paga_spatial_simulation(
    eykthyr: Eykthyr,
    TFs: Sequence[str],
    cluster_name: str,
):
    """"""
    umap_spatial_simulation(
        eykthyr,
        TFs,
        [eykthyr.perturbed_X],
        cluster_name=cluster_name,
    )


def prep_paga(
    eykthyr: Eykthyr,
    groups: str,
):
    sc.pp.neighbors(eykthyr.perturbed_X, use_rep="normalized_X")
    sc.tl.umap(eykthyr.perturbed_X)
    sc.tl.paga(eykthyr.perturbed_X, groups=groups)
    sc.tl.draw_graph(eykthyr.perturbed_X, init_pos="X_umap", random_state=123)
    sc.pl.draw_graph(eykthyr.perturbed_X, color=groups, legend_loc="on data")


def umap_spatial_simulation(
    eykthyr,
    TFs,
    datasets,
    n_grid=40,
    min_masses=[0.27, 0.007],
    scales=[30, 1.2],
    embeddings=["spatial", "X_draw_graph_fr"],
    n_neighbors=[20, 25],
    cluster_name="original_leiden",
    show_plots=[True, True],
):

    eykthyr.RNA.X = eykthyr.RNA.X.astype("double")

    for TF in TFs:
        for embedding in embeddings:
            eykthyr.embeddings[embedding] = Embedding()
        eykthyr.estimate_transition_probs(
            embedding_names=embeddings,
            tf_name=TF,
            n_neighbors=n_neighbors,
            sampled_fraction=1,
        )
        eykthyr.calculate_embedding_shifts(embedding_names=embeddings, sigma_corr=0.05)
        eykthyr.calculate_p_mass(
            embeddings,
            smooth=0.8,
            n_grid=n_grid,
            n_neighbors=n_neighbors,
        )
        eykthyr.calculate_mass_filter(embeddings, min_mass=min_masses, plot=False)
        # eykthyr.suggest_mass_thresholds(n_suggestion=12)
        for embedding, show_plot, scale in zip(embeddings, show_plots, scales):
            if show_plot == True:
                fig, ax = plt.subplots(1, 2, figsize=[13, 6])

                eykthyr.plot_simulation_flow_on_grid(embedding, scale=scale, ax=ax[0])
                ax[0].set_title(f"Simulated cell identity shift vector: {TF} KO")

                # Show quiver plot that was calculated with randomized graph.
                eykthyr.plot_simulation_flow_random_on_grid(
                    embedding,
                    scale=scale,
                    ax=ax[1],
                )
                ax[1].set_title(f"Randomized simulation vector")

                plt.show()

        fig2, ax2 = plt.subplots(1, len(embeddings), figsize=[16, 6])

        for i, embedding in enumerate(embeddings):

            eykthyr.plot_cluster_whole(embedding, cluster_name, ax=ax2[i], s=5)
            eykthyr.plot_simulation_flow_on_grid(
                embedding,
                scale=scales[i],
                ax=ax2[i],
                show_background=False,
            )
            ax2[i].set_title(f"Simulated cell identity shift {embedding}: {TF} KO")

        plt.show()


"""
def old_umap_spatial_simulation(TFs,
    datasets,
    n_grid=40,
    min_masses=[.27, 0.007],
    scales=[30,1.2],
    embeddings=['spatial','X_draw_graph_fa'],
    n_neighbors=[20,25],
    cluster_name='original_leiden',
    show_plots=[True, True]
):

    for TF in TFs:
        oracless = []
        for embedding, n in zip(embeddings, n_neighbors):
            oracles = []
            for d in datasets:

                oracle = co.Oracle()
                oracle.import_anndata_as_raw_count(adata=d,
                                   cluster_column_name=cluster_name,
                                   embedding_name=embedding)
                oracles.append(oracle)
            oracless.append(oracles)


            for oracle in oracles:
                oracle.adata.X = oracle.adata.X.astype('double')
                oracle.adata.layers['imputed_count'] = oracle.adata.obsm[f'normalized_X_{TF}_dropout'].astype('double')
                oracle.adata.layers['delta_X'] =  oracle.adata.obsm[
                                        f'normalized_X_{TF}_dropout'] - oracle.adata.obsm['normalized_X']
                oracle.estimate_transition_prob(
                                n_neighbors=n,
                                sampled_fraction=1
                                )
                oracle.calculate_embedding_shift(sigma_corr=0.05)



        for i in range(1):
            for j,oracles in enumerate(oracless):


                goi = TF
                # Show quiver plot

                oracles[i].calculate_p_mass(smooth=0.8, n_grid=n_grid, n_neighbors=n_neighbors[j])
#                 oracles[i].suggest_mass_thresholds(n_suggestion=12)
                oracles[i].calculate_mass_filter(min_mass=min_masses[j], plot=False)
                if show_plots[j] == True:
                    fig, ax = plt.subplots(1, 2,  figsize=[13, 6])

                    oracles[i].plot_simulation_flow_on_grid(scale=scales[j], ax=ax[0])
                    ax[0].set_title(f"Simulated cell identity shift vector: {goi} KO")

                    # Show quiver plot that was calculated with randomized graph.
                    oracles[i].plot_simulation_flow_random_on_grid(scale=scales[j], ax=ax[1])
                    ax[1].set_title(f"Randomized simulation vector")

                    plt.show()

        # Plot vector field with cell cluster
            fig2, ax2 = plt.subplots(1,len(embeddings),figsize=[16, 6])
            for oracles, embedding, j in zip(oracless, embeddings, range(len(oracless))):

                oracles[i].plot_cluster_whole(ax=ax2[j], s=5)
                oracles[i].plot_simulation_flow_on_grid(scale=scales[j], ax=ax2[j], show_background=False)
                ax2[j].set_title(f"Simulated cell identity shift {embedding}: {goi} KO")

            plt.show()

def development_simulation():

    from celloracle.applications import Gradient_calculator
    from celloracle.applications import Oracle_development_module


    return
"""



from dataclasses import dataclass

import numpy as np
from scipy.sparse import spmatrix


@dataclass
class Embedding:

    embedding: np.ndarray = None
    embedding_knn: spmatrix = None
    sampling_ixs: np.ndarray = None
    corrcoef: np.ndarray = None
    corrcoef_random: np.ndarray = None
    transition_prob: np.ndarray = None
    transition_prob_random: np.ndarray = None
    delta_embedding: np.ndarray = None
    delta_embedding_random: np.ndarray = None
    total_p_mass: np.ndarray = None
    flow_embedding: np.ndarray = None
    flow_grid: np.ndarray = None
    flow: np.ndarray = None
    flow_norm: np.ndarray = None
    flow_norm_magnitude: np.ndarray = None
    flow_rndm: np.ndarray = None
    flow_norm_rndm: np.ndarray = None
    flow_norm_magnitude_rndm: np.ndarray = None
    min_mass: float = None
    mass_filter: np.ndarray = None
    colorandum: np.ndarray = None



from pathlib import Path
from typing import Dict, Optional, Sequence, Union

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import torch
from popari import pl, tl
from popari.components import PopariDataset
from popari.io import save_anndata
from popari.model import Popari, load_trained_model

from .embedding import Embedding
from .modified_VelocytoLoom_class import modified_VelocytoLoom
from .util import get_metagene_edges_window, run_all_perturbations


class Eykthyr(modified_VelocytoLoom):

    def __init__(
        self,
        RNA: Optional[sc.AnnData] = None,
        ATAC: Optional[sc.AnnData] = None,
        popari: Optional[Popari] = None,
        TF: Optional[sc.AnnData] = None,
        edge_weights: Optional[sc.AnnData] = None,
        perturbed_X: Optional[sc.AnnData] = None,
        name: str = "eykthyr_dataset",
        cluster_annotation: Sequence[str] = [],
        num_metagenes: int = -1,
        embeddings: Optional[Dict[str, Embedding]] = {},
    ):

        self.RNA = RNA
        self.ATAC = ATAC
        self.popari = popari
        self.TF = TF
        self.edge_weights = edge_weights
        self.perturbed_X = perturbed_X

        self.datasetname = name
        self.rna_preprocessed = False

        self.cluster_annotation = cluster_annotation
        self.num_metagenes = num_metagenes
        self.embeddings = embeddings

    def preprocess_rna(
        self,
        make_plots: bool = False,
        cluster_annotation: Optional[Sequence[str]] = [],
    ):
        """"""

        if self.rna_preprocessed == True:
            print("RNA appears to already be preprocessed, doing nothing")
            return

        sc.pp.filter_genes(self.RNA, min_cells=5)
        sc.pp.filter_cells(self.RNA, min_counts=10)
        self.RNA.layers["raw"] = self.RNA.X
        sc.pp.normalize_total(self.RNA, inplace=True, target_sum=10000)
        sc.pp.log1p(self.RNA)
        if len(cluster_annotation) > 0:
            self.cluster_annotation = cluster_annotation
        if make_plots == True:
            sc.pp.pca(self.RNA)
            sc.pp.neighbors(self.RNA)
            sc.tl.umap(self.RNA)
            sc.tl.leiden(self.RNA, key_added="clusters")
            sc.pl.umap(self.RNA, color=["clusters"] + self.cluster_annotation)
        self.rna_preprocessed = True

    def compute_metagenes(
        self,
        K: int = 16,
        lambda_Sigma_x_inv: float = 1e-4,
        torch_context: dict = dict(device="cuda:0", dtype=torch.float64),
        initial_iterations: int = 10,
        spatial_iterations: int = 200,
    ):
        """"""

        if self.rna_preprocessed == False:
            print(
                "RNA appears to not be preprocessed. Please preprocess RNA using Eykthyr.preprocess_rna() or set Eykthyr.rna_preprocessed = True",
            )
            return

        self.num_metagenes = K
        popari_d = PopariDataset(self.RNA, self.datasetname)
        popari_d.compute_spatial_neighbors()
        self.RNA.obs["adjacency_list"] = popari_d.obs["adjacency_list"]
        self.RNA.obsp["adjacency_matrix"] = popari_d.obsp["adjacency_matrix"]
        self.RNA.X = self.RNA.X.todense()

        self.popari = Popari(
            K=K,
            replicate_names=self.datasetname,
            datasets=[self.RNA],
            lambda_Sigma_x_inv=lambda_Sigma_x_inv,
            torch_context=torch_context,
            initial_context=torch_context,
            verbose=0,
        )

        for iteration in range(initial_iterations):
            self.popari.estimate_parameters(update_spatial_affinities=False)
            self.popari.estimate_weights(use_neighbors=False)

        for iteration in range(spatial_iterations):
            self.popari.estimate_parameters()
            self.popari.estimate_weights()

    def analyze_metagenes(
        self,
        num_leiden_clusters: int = 10,
    ):
        """"""

        if not self.popari:
            print("Popari has not been run. Please run compute_metagenes() first.")
            return
        tl.preprocess_embeddings(self.popari)
        tl.leiden(
            self.popari,
            use_rep="normalized_X",
            target_clusters=num_leiden_clusters,
        )
        sc.pp.neighbors(
            self.popari.datasets[0],
            use_rep="normalized_X",
            key_added="norm_X_neighbors",
        )
        sc.tl.umap(self.popari.datasets[0], neighbors_key="norm_X_neighbors")
        sc.pl.umap(self.popari.datasets[0], color=["leiden"] + self.cluster_annotation)

    def save_anndata(
        self,
        dirpath: str,
    ):
        """"""

        dirpath = Path(dirpath)
        path_without_extension = dirpath.parent / dirpath.stem
        path_without_extension.mkdir(exist_ok=True)
        if self.RNA:
            self.RNA.uns["datasetname"] = self.datasetname
            self.RNA.uns["rna_preprocessed"] = self.rna_preprocessed
            self.RNA.uns["cluster_annotation"] = self.cluster_annotation
            self.RNA.uns["num_metagenes"] = self.num_metagenes
            if "adjacency_list" in self.RNA.obs.columns:
                del self.RNA.obs["adjacency_list"]
            self.RNA.write(f"{path_without_extension}/RNA.h5ad")
        if self.ATAC:
            self.ATAC.write(f"{path_without_extension}/ATAC.h5ad")
        if self.popari:
            self.popari.save_results(f"{path_without_extension}/popari.h5ad")
        if self.TF:
            self.TF.write(f"{path_without_extension}/TF.h5ad")
        if self.edge_weights:
            self.edge_weights.write(f"{path_without_extension}/edge_weights.h5ad")
        if self.perturbed_X:
            del self.perturbed_X.obs["adjacency_list"]
            self.perturbed_X.write(f"{path_without_extension}/perturbed_X.h5ad")

    def set_RNA(
        self,
        RNA: sc.AnnData,
    ):
        """"""

        self.RNA = RNA

    def set_ATAC(
        self,
        ATAC: sc.AnnData,
    ):
        """"""

        self.ATAC = ATAC

    def set_popari(
        self,
        popari: Popari,
    ):
        """"""

        self.popari = popari

    def set_TF(
        self,
        TF: sc.AnnData,
    ):
        """"""

        self.TF = TF

    def set_edge_weights(
        self,
        edge_weights: sc.AnnData,
    ):
        """"""

        self.edge_weights = edge_weights

    def set_perturbed_X(
        self,
        perturbed_X: sc.AnnData,
    ):
        """"""

        self.perturbed_X = perturbed_X

    def compute_TF_activity(
        self,
        peak_tsv: str,
        archr_dataset_name: str,
        motif_tsv: str,
    ):
        """"""

        archr_name_len = len(archr_dataset_name) + 1
        tfpeaks = pd.read_csv(peak_tsv, sep=" ")
        new_col_names = [c[archr_name_len:-2] for c in tfpeaks.columns]
        tfpeaks.rename(
            columns={c: new_c for c, new_c in zip(tfpeaks.columns, new_col_names)},
            inplace=True,
        )
        tfpeaks = tfpeaks.T

        tfmotifs = pd.read_csv(motif_tsv, sep=" ")
        tfmotifs.index = range(1, tfmotifs.shape[0] + 1)
        tfmotifs = tfmotifs.rename(
            columns={c: c.split("_")[0] for c in tfmotifs.columns},
        )
        tfmotifs = tfmotifs.astype(int)

        cellmotifs = sc.AnnData(tfpeaks.dot(tfmotifs))
        cellmotifs.layers["raw"] = cellmotifs.X

        sc.pp.normalize_total(cellmotifs, exclude_highly_expressed=True)
        sc.pp.scale(cellmotifs, zero_center=False)

        included = self.RNA.obs.index[self.RNA.obs.index.isin(cellmotifs.obs.index)]
        excluded = self.RNA.obs.index[~self.RNA.obs.index.isin(cellmotifs.obs.index)]
        cellmotifs = cellmotifs[included, :]
        tf_archr = cellmotifs.to_df()
        for exclude in excluded:
            tf_archr.loc[exclude, :] = tf_archr.mean(axis=0)
        tf_archr = tf_archr.rename(
            columns={c: c.split("_")[0] for c in tf_archr.columns},
        )
        tf_archr = tf_archr.loc[:, ~tf_archr.columns.duplicated()]
        self.TF = sc.AnnData(tf_archr)
        self.TF = self.TF[self.RNA.obs_names, :]
        self.TF.obsm["spatial"] = self.RNA.obsm["spatial"]
        self.TF.layers["raw"] = self.TF.X
        self.TF.X -= self.TF.X.min(axis=0)
        self.TF.X /= self.TF.X.max(axis=0)

    def compute_TF_metagene_weights(
        self,
        num_hops: int = 2,
    ):
        """"""

        if not self.popari or not self.TF:
            print(
                "Popari needs to be run first, please run compute_metagenes().\n"
                "TF activity also needs to be computed by ArchR. Please follow the jupyter notebook for\n"
                "preprocessing ATAC-seq data and then run compute_TF_activity().",
            )
            return

        M_edges = []
        for j in range(self.num_metagenes):
            temp = []
            edges = get_metagene_edges_window(
                self.popari.datasets[0],
                self.TF,
                j,
                self.popari.datasets[0],
                num_hops=num_hops,
            )
            Madata = sc.AnnData(edges)
            Madata = Madata[self.popari.datasets[0].obs_names, :]
            Madata.obsm["spatial"] = self.popari.datasets[0].obsm["spatial"]
            temp.append(Madata)
            M_edges.append(temp)
        self.edge_weights = M_edges[0][0]
        for i in range(self.num_metagenes):
            self.edge_weights.layers[f"M_{i}"] = M_edges[i][0].X

    def run_all_perturbations(
        self,
    ):
        """"""

        if not self.popari:
            print("Popari needs to be run first, please run compute_metagenes().")
            return

        if not self.TF:
            print(
                "TF activity needs to be computed by ArchR. Please follow the jupyter notebook for\n"
                "preprocessing ATAC-seq data and then run compute_TF_activity().",
            )
            return

        if not self.edge_weights:
            print(
                "GRN edge weights need to be computed first, please run compute_TF_metagene_weights().",
            )
            return

        for d in self.popari.datasets:
            d.obs["original_leiden"] = d.obs["leiden"]
        run_all_perturbations(
            self.popari,
            [self.TF],
            [self.edge_weights],
            K=self.num_metagenes,
            useX=True,
        )
        self.perturbed_X = sc.AnnData(
            X=self.popari.datasets[0].obsm["X"],
            obs=self.popari.datasets[0].obs,
            obsm=self.popari.datasets[0].obsm,
        )


def load_anndata(dirpath):
    """"""

    dirpath = Path(dirpath)
    path_without_extension = dirpath.parent / dirpath.stem
    RNA = ATAC = TF = edge_weights = perturbed_X = popari = None
    if Path(f"{path_without_extension}/RNA.h5ad").is_file():
        RNA = sc.read(f"{path_without_extension}/RNA.h5ad")
    if Path(f"{path_without_extension}/ATAC.h5ad").is_file():
        ATAC = sc.read(f"{path_without_extension}/ATAC.h5ad")
    if Path(f"{path_without_extension}/TF.h5ad").is_file():
        TF = sc.read(f"{path_without_extension}/TF.h5ad")
    if Path(f"{path_without_extension}/edge_weights.h5ad").is_file():
        edge_weights = sc.read(f"{path_without_extension}/edge_weights.h5ad")
    if Path(f"{path_without_extension}/perturbed_X.h5ad").is_file():
        perturbed_X = sc.read(f"{path_without_extension}/perturbed_X.h5ad")
    if Path(f"{path_without_extension}/popari.h5ad").is_file():
        popari = load_trained_model(f"{path_without_extension}/popari.h5ad")
    elif Path(f"{path_without_extension}/popari").is_dir():
        popari = load_trained_model(f"{path_without_extension}/popari.h5ad")
    eykthyr = Eykthyr(
        RNA=RNA,
        ATAC=ATAC,
        TF=TF,
        edge_weights=edge_weights,
        perturbed_X=perturbed_X,
        popari=popari,
    )
    # eykthyr.datasetname = eykthyr.RNA.uns['datasetname']
    # eykthyr.rna_preprocessed = eykthyr.RNA.uns['rna_preprocessed']
    # eykthyr.cluster_annotation = eykthyr.RNA.uns['cluster_annotation']
    # eykthyr.num_metagenes = eykthyr.RNA.uns['num_metagenes']
    return eykthyr
